%------------------------------------------------------------------------------%
% mexpat.xml.m
% Peter Ross <pro@missioncriticalit.com>
%
% A generic XML representation that can be created with the mexpat parser.
%
%------------------------------------------------------------------------------%

/****M* libs/mexpat/mexpat.xml.m/mexpat__xml
 * NAME
 *   mexpat__xml
 * PURPOSE
 *   <DOCXXX:module description>
 *         
 * AUTHOR
 *   <DOCXXX:list of authors>
 * HISTORY
 *   $Id: mexpat.xml.m,v 1.11 2005/06/09 11:13:35 winx820 Exp $ (use cvs log to get detailed history)
 * BUGS
 *   <bug descriptions>
 * TODO
 *   <list of todos>
 ****
 * <anything you want that won't be in the doc>
 */
 :- module mexpat__xml.

:- interface.

:- import_module io, list.

:- typeclass xmlize(T) where [
    func to_xml(T) = xml.element,
    func from_xml(xml.element) = T
].
:- instance xmlize(int).
:- instance xmlize(string).
:- instance xmlize(list(T)) <= xmlize(T).

    % A simple representation of XML documents.
/****T* libs/mexpat/mexpat.xml.m/element
 * NAME
 *    Type: element
 * DESCRIPTION
 *    <DOCXXX: description of type element>
 ****
 * <anything you want that won't be in the doc>
 */
:- type element
        --->    elem(
                    name    :: string,
                    attrs   :: list(attribute),
                    content :: list(content)
                ).

/****T* libs/mexpat/mexpat.xml.m/attribute
 * NAME
 *    Type: attribute
 * DESCRIPTION
 *    <DOCXXX: description of type attribute>
 ****
 * <anything you want that won't be in the doc>
 */
:- type attribute
        --->    attr(
                    attr_name   :: string,
                    attr_value  :: string
                ).

/****T* libs/mexpat/mexpat.xml.m/content
 * NAME
 *    Type: content
 * DESCRIPTION
 *    <DOCXXX: description of type content>
 ****
 * <anything you want that won't be in the doc>
 */
:- type content
        --->    celem(element)
        ;       ctext(string)
        .

    % Return child elements of the current element.
/****F* libs/mexpat/mexpat.xml.m/child_elements
 * NAME
 *    child_elements
 * FUNCTION
 *    <DOCXXX: description for child_elements>
 * INPUTS
 *    <DOCXXX: inputs types  (and modes if not "in" ) and description>
 * OUTPUT
 *    <DOCXXX: outputs types and description>
 * TODO
 *    <known todos for this function>
 ****
 * <anything you want that won't be in the doc>
 */
 :- func child_elements(element) = list(element).

    % Return attribute value if it exists.
/****F* libs/mexpat/mexpat.xml.m/attr
 * NAME
 *    attr
 * FUNCTION
 *    <DOCXXX: description for attr>
 * INPUTS
 *    <DOCXXX: inputs types  (and modes if not "in" ) and description>
 * OUTPUT
 *    <DOCXXX: outputs types and description>
 * TODO
 *    <known todos for this function>
 ****
 * <anything you want that won't be in the doc>
 */
 :- func attr(string::in, list(attribute)::in) = (string::out) is semidet.

    % Return the attribute value if it exists, otherwise throw exception.
/****F* libs/mexpat/mexpat.xml.m/attribute
 * NAME
 *    attribute
 * FUNCTION
 *    <DOCXXX: description for attribute>
 * INPUTS
 *    <DOCXXX: inputs types  (and modes if not "in" ) and description>
 * OUTPUT
 *    <DOCXXX: outputs types and description>
 * TODO
 *    <known todos for this function>
 ****
 * <anything you want that won't be in the doc>
 */
 :- func attribute(string, list(attribute)) = string.


    % Give some content return the text,
    % otherwise throw an exception.
/****F* libs/mexpat/mexpat.xml.m/txt
 * NAME
 *    txt
 * FUNCTION
 *    <DOCXXX: description for txt>
 * INPUTS
 *    <DOCXXX: inputs types  (and modes if not "in" ) and description>
 * OUTPUT
 *    <DOCXXX: outputs types and description>
 * TODO
 *    <known todos for this function>
 ****
 * <anything you want that won't be in the doc>
 */
 :- func txt(content) = string.

    % Give some content return the element,
    % otherwise throw an exception.
/****F* libs/mexpat/mexpat.xml.m/el
 * NAME
 *    el
 * FUNCTION
 *    <DOCXXX: description for el>
 * INPUTS
 *    <DOCXXX: inputs types  (and modes if not "in" ) and description>
 * OUTPUT
 *    <DOCXXX: outputs types and description>
 * TODO
 *    <known todos for this function>
 ****
 * <anything you want that won't be in the doc>
 */
 :- func el(content) = element.


%------------------------------------------------------------------------------%

/****T* libs/mexpat/mexpat.xml.m/maybe_element
 * NAME
 *    Type: maybe_element
 * DESCRIPTION
 *    <DOCXXX: description of type maybe_element>
 ****
 * <anything you want that won't be in the doc>
 */
:- type maybe_element
        --->    element(element)
        ;       error(
                    line    :: int,
                    column  :: int,
                    message :: string
                ).

    %
    % Given a string which represents a complete XML document,
    % return the element structure which represents that document.
    %
/****F* libs/mexpat/mexpat.xml.m/parse_string
 * NAME
 *    parse_string
 * FUNCTION
 *    <DOCXXX: description for parse_string>
 * INPUTS
 *    <DOCXXX: inputs types  (and modes if not "in" ) and description>
 * OUTPUT
 *    <DOCXXX: outputs types and description>
 * TODO
 *    <known todos for this function>
 ****
 * <anything you want that won't be in the doc>
 */
 :- func parse_string(string) = maybe_element.

%------------------------------------------------------------------------------%
%
% Combinators which can be used to select content located in the 
% content.
%
% These combinators are the same as those from HaXML.
% See the HaXML documentation for the documentation of what these do.
%


:- type none          ---> none.
:- type keep          ---> keep.  
:- type elem          ---> elem.  
:- type text          ---> text.  
:- type children      ---> children.
:- type tag           ---> tag(string).
:- type attr          ---> attr(string).
:- type attrval       ---> attrval(attribute).
:- type o(T, U)       ---> T `o` U.
:- type append(T, U)  ---> T `append` U.
:- type with(T, U)    ---> T `with` U.
:- type without(T, U) ---> T `without` U.
:- type cat(T)        ---> cat(list(T)).
:- type ife(C, T, E)  ---> ife(C, T, E).
:- type deep(T)       ---> deep(T).
:- type deepest(T)    ---> deepest(T).
:- type multi(T)      ---> multi(T).
:- type chip(T)       ---> chip(T).
:- type foldXml(T)    ---> foldXml(T).

:- type F `'/>'` G    ==   G `o` children `o` F.
:- type F `'</'` G    ==   F `with` (G `o` children).
:- type F `'|>|'` G   ==   ife(F, F, G).

:- func F `'/>'` G = F `'/>'` G.
:- func F `'</'` G = F `'</'` G.
:- func F `'|>|'` G = F `'|>|'` G.

:- typeclass cfilter(T) where [
    func cfilter(T, content) = list(content)
].

:- instance cfilter(none).
:- instance cfilter(keep).
:- instance cfilter(elem).
:- instance cfilter(text).
:- instance cfilter(children).
:- instance cfilter(tag).
:- instance cfilter(xml.attr).
:- instance cfilter(attrval).
:- instance cfilter(o(T, U)) <= (cfilter(T), cfilter(U)).
:- instance cfilter(append(T, U)) <= (cfilter(T), cfilter(U)).
:- instance cfilter(with(T, U)) <= (cfilter(T), cfilter(U)).
:- instance cfilter(without(T, U)) <= (cfilter(T), cfilter(U)).
:- instance cfilter(cat(T)) <= cfilter(T).
:- instance cfilter(ife(C, T, E)) <= (cfilter(C), cfilter(T), cfilter(E)).
:- instance cfilter(deep(T)) <= cfilter(T).
:- instance cfilter(deepest(T)) <= cfilter(T).
:- instance cfilter(multi(T)) <= cfilter(T).
:- instance cfilter(chip(T)) <= cfilter(T).
:- instance cfilter(foldXml(T)) <= cfilter(T).

:- typeclass path(T) where [
    func path(T, path) = list(path)
].

:- type path == list(content).

:- instance path(none).
:- instance path(keep).
:- instance path(elem).
:- instance path(text).
:- instance path(children).
:- instance path(tag).
:- instance path(xml.attr).
:- instance path(attrval).
:- instance path(o(T, U)) <= (path(T), path(U)).
:- instance path(append(T, U)) <= (path(T), path(U)).
:- instance path(with(T, U)) <= (path(T), path(U)).
:- instance path(without(T, U)) <= (path(T), path(U)).
:- instance path(cat(T)) <= path(T).
:- instance path(ife(C, T, E)) <= (path(C), path(T), path(E)).
:- instance path(deep(T)) <= path(T).
:- instance path(deepest(T)) <= path(T).
:- instance path(multi(T)) <= path(T).

%------------------------------------------------------------------------------%

    %
    % Write the generic xml representation as an XML document to the current
    % output stream, or a specified output stream.
    %
/****P* libs/mexpat/mexpat.xml.m/write_element 
 * NAME
 *    write_element 
 * SYNOPSIS
 *    <DOCXXX:short descr for write_element >
 * FUNCTION
 *    <DOCXXX:description for write_element >
 * PARAMETERS
 *    <DOCXXX: types and modes and description for write_element>
 * RESULT
 *    <DOCXXX: determinism for write_element>
 * TODO
 *    <known todos for this function>
 ****
 * <anything you want that won't be in the doc>
 */
:- pred write_element(element::in, io::di, io::uo) is det.
/****P* libs/mexpat/mexpat.xml.m/write_element 
 * NAME
 *    write_element 
 * SYNOPSIS
 *    <DOCXXX:short descr for write_element >
 * FUNCTION
 *    <DOCXXX:description for write_element >
 * PARAMETERS
 *    <DOCXXX: types and modes and description for write_element>
 * RESULT
 *    <DOCXXX: determinism for write_element>
 * TODO
 *    <known todos for this function>
 ****
 * <anything you want that won't be in the doc>
 */
:- pred write_element(io__output_stream::in, element::in,
                io::di, io::uo) is det.

%------------------------------------------------------------------------------%

/****T* libs/mexpat/mexpat.xml.m/parser.
 * NAME
 *    Type: parser.
 * DESCRIPTION
 *    <DOCXXX: description of type parser.>
 ****
 * <anything you want that won't be in the doc>
 */
:- type parser.
/****T* libs/mexpat/mexpat.xml.m/parser
 * NAME
 *    Instance: parser
 * DESCRIPTION
 *    <DOCXXX: description of instance parser>
 * DERIVED FROM
 *    An instance of the typeclass mexpat
 ****
 * <anything you want that won't be in the doc>
 */
:- instance mexpat(parser).

/****F* libs/mexpat/mexpat.xml.m/parser_init
 * NAME
 *    parser_init
 * FUNCTION
 *    <DOCXXX: description for parser_init>
 * INPUTS
 *    <DOCXXX: inputs types  (and modes if not "in" ) and description>
 * OUTPUT
 *    <DOCXXX: outputs types and description>
 * TODO
 *    <known todos for this function>
 ****
 * <anything you want that won't be in the doc>
 */
 :- func parser_init = parser.
/****P* libs/mexpat/mexpat.xml.m/get_element 
 * NAME
 *    get_element 
 * SYNOPSIS
 *    <DOCXXX:short descr for get_element >
 * FUNCTION
 *    <DOCXXX:description for get_element >
 * PARAMETERS
 *    <DOCXXX: types and modes and description for get_element>
 * RESULT
 *    <DOCXXX: determinism for get_element>
 * TODO
 *    <known todos for this function>
 ****
 * <anything you want that won't be in the doc>
 */
:- pred get_element(parser::in, element::out) is semidet.

/****P* libs/mexpat/mexpat.xml.m/serialize 
 * NAME
 *    serialize 
 * SYNOPSIS
 *    Turn an element representation into it's string
 *    representation.
 * SOURCE
 */
:- func serialize(element) = string.
/*****/

%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%

:- implementation.

:- import_module require.

parse_string(String) = Xml :-
    GenericXMLParser = parser_init,
    Parser0 = init(GenericXMLParser),

    xml_parse(String, yes, ParseResult, Parser0, Parser),

    ( ParseResult = ok,
        (
            get_user_data(Parser, UserData),
            get_element(UserData, Element)
        ->
            Xml = element(Element)
        ;
            error("mexpat__xml__parse_string: Unable to extract element.")
        )
    ; ParseResult = error(Line, Column, Message),
        Xml = error(Line, Column, Message)
    ).

%------------------------------------------------------------------------------%

F `'/>'` G = G `o` children `o` F.
F `'</'` G = F `with` (G `o` children).
F `'|>|'` G = ife(F, F, G).

%------------------------------------------------------------------------------%

:- instance path(none) where [
    path(none, _) = []
].

:- instance path(keep) where [
    path(keep, P) = [P]
].

:- instance path(elem) where [
    path(elem, []) = [],
    path(elem, [ctext(_) | _]) = [],
    path(elem, P @ [celem(_) | _]) = [P]
].

:- instance path(text) where [
    path(text, []) = [],
    path(text, P @ [ctext(_) | _]) = [P],
    path(text, [celem(_) | _]) = []
].

:- instance path(children) where [
    path(children, []) = [],
    path(children, [ctext(_) | _]) = [],
    path(children, P @ [celem(E) | _]) = map(func(C) = [C | P], E ^ content)
].

:- instance path(tag) where [
    path(tag(_), []) = [],
    path(tag(_), [ctext(_) | _]) = [],
    path(tag(T), P @ [celem(E) | _]) = ( E ^ name = T -> [P] ; [] )
].

:- instance path(xml.attr) where [
    path(attr(_), []) = [],
    path(attr(_), [ctext(_) | _]) = [],
    path(attr(A), P @ [celem(E) | _]) = ( find_attribute(A, E ^ attrs, _) -> [P] ; [] )
].

:- instance path(attrval) where [
    path(attrval(_), []) = [],
    path(attrval(_), [ctext(_) | _]) = [],
    path(attrval(A), P @ [celem(E) | _]) = ( find_attribute(A ^ attr_name, E ^ attrs, A) -> [P] ; [] )
].

:- instance path(o(T, U)) <= (path(T), path(U)) where [
    path(F `o` G, P) = condense(map(path(F), path(G, P)))
].

:- instance path(append(T, U)) <= (path(T), path(U)) where [
    path(F `append` G, P) = path(F, P) ++ path(G, P)
].

:- instance path(with(T, U)) <= (path(T), path(U)) where [
    path(F `with` G, P) = filter(nonempty_path(G), path(F, P))
].

:- instance path(without(T, U)) <= (path(T), path(U)) where [
    path(F `without` G, P) = filter(empty_path(G), path(F, P))
].

:- instance path(cat(T)) <= path(T) where [
    path(cat(Fs), P) = condense(map(func(F) = path(F, P), Fs))
].

:- instance path(ife(C, T, E)) <= (path(C), path(T), path(E)) where [
    path(ife(C, T, E), P) = ( nonempty_path(C, P) -> path(T, P) ; path(E, P) )
].

:- instance path(deep(T)) <= path(T) where [
    path(deep(F), P) = path(F `'|>|'` ife(children, deep(F `o` children), none), P)
].

:- instance path(deepest(T)) <= path(T) where [
    path(deepest(F), P) = path(ife(F `o` children, deepest(F `o` children), none) `'|>|'` F, P)
].

:- instance path(multi(T)) <= path(T) where [
    path(multi(F), P) = path(F `append` ife(F `o` children, multi(F `o` children), none), P)
].

:- pred nonempty_path(P::in, path::in) is semidet <= path(P).
nonempty_path(F, C) :-
    not empty_path(F, C).

:- pred empty_path(P::in, path::in) is semidet <= path(P).
empty_path(F, C) :-
    path(F, C) = [].

%------------------------------------------------------------------------------%

:- instance cfilter(none) where [
    cfilter(none, _) = []
].

:- instance cfilter(keep) where [
    cfilter(keep, P) = [P]
].

:- instance cfilter(elem) where [
    cfilter(elem, ctext(_)) = [],
    cfilter(elem, P @ celem(_)) = [P]
].

:- instance cfilter(text) where [
    cfilter(text, P @ ctext(_)) = [P],
    cfilter(text, celem(_)) = []
].

:- instance cfilter(children) where [
    cfilter(children, ctext(_)) = [],
    cfilter(children, celem(E)) = E ^ content
].

:- instance cfilter(tag) where [
    cfilter(tag(_), ctext(_)) = [],
    cfilter(tag(T), P @ celem(E)) = ( E ^ name = T -> [P] ; [] )
].

:- instance cfilter(xml.attr) where [
    cfilter(attr(_), ctext(_)) = [],
    cfilter(attr(A), P @ celem(E)) = ( find_attribute(A, E ^ attrs, _) -> [P] ; [] )
].

:- instance cfilter(attrval) where [
    cfilter(attrval(_), ctext(_)) = [],
    cfilter(attrval(A), P @ celem(E)) = ( find_attribute(A ^ attr_name, E ^ attrs, A) -> [P] ; [] )
].

:- instance cfilter(o(T, U)) <= (cfilter(T), cfilter(U)) where [
    cfilter(F `o` G, P) = condense(map(cfilter(F), cfilter(G, P)))
].

:- instance cfilter(append(T, U)) <= (cfilter(T), cfilter(U)) where [
    cfilter(F `append` G, P) = cfilter(F, P) ++ cfilter(G, P)
].

:- instance cfilter(with(T, U)) <= (cfilter(T), cfilter(U)) where [
    cfilter(F `with` G, P) = filter(nonempty_cfilter(G), cfilter(F, P))
].

:- instance cfilter(without(T, U)) <= (cfilter(T), cfilter(U)) where [
    cfilter(F `without` G, P) = filter(empty_cfilter(G), cfilter(F, P))
].

:- instance cfilter(cat(T)) <= cfilter(T) where [
    cfilter(cat(Fs), P) = condense(map(func(F) = cfilter(F, P), Fs))
].

:- instance cfilter(ife(C, T, E)) <= (cfilter(C), cfilter(T), cfilter(E)) where [
    cfilter(ife(C, T, E), P) = ( nonempty_cfilter(C, P) -> cfilter(T, P) ; cfilter(E, P) )
].

:- instance cfilter(deep(T)) <= cfilter(T) where [
    cfilter(deep(F), P) = cfilter(F `'|>|'` ife(children, deep(F `o` children), none), P)
].

:- instance cfilter(deepest(T)) <= cfilter(T) where [
    cfilter(deepest(F), P) = cfilter(ife(F `o` children, deepest(F `o` children), none) `'|>|'` F, P)
].

:- instance cfilter(multi(T)) <= cfilter(T) where [
    cfilter(multi(F), P) = cfilter(F `append` ife(F `o` children, multi(F `o` children), none), P)
].

:- instance cfilter(chip(T)) <= cfilter(T) where [
    cfilter(chip(F), P) =
        ( P = celem(elem(N, As, _)) ->
            [celem(elem(N, As, cfilter(F `o` children, P)))]
        ;
            [P]
        )
].

:- instance cfilter(foldXml(T)) <= cfilter(T) where [
    cfilter(foldXml(F), P) = cfilter(F `o` chip(foldXml(F)), P)
].


:- pred nonempty_cfilter(P::in, content::in) is semidet <= cfilter(P).
nonempty_cfilter(F, C) :-
    not empty_cfilter(F, C).

:- pred empty_cfilter(P::in, content::in) is semidet <= cfilter(P).
empty_cfilter(F, C) :-
    cfilter(F, C) = [].

%------------------------------------------------------------------------------%

:- type cfilter
    --->    none
    ;       keep
    ;       elem
    ;       text
    ;       children

    ;       tag(string)
    ;       attr(string)
    ;       attrval(attribute)

    ;       cfilter `o` cfilter
    ;       cfilter ++ cfilter
    ;       cfilter `with` cfilter
    ;       cfilter `without` cfilter
    ;       cfilter `'/>'` cfilter
    ;       cfilter `'</'` cfilter
    ;       cfilter `'|>|'` cfilter

    ;       cat(list(cfilter))

    ;       ife(cfilter, cfilter, cfilter)

    ;       deep(cfilter)
    ;       deepest(cfilter)
    ;       multi(cfilter)

    ;       chip(cfilter)
    ;       foldXml(cfilter)
    .

    %
    % Apply the cfilter to a piece of content getting the filtered content
    % back.
    %
:- func apply_cfilter(cfilter, content) = list(content).

apply_cfilter(none, _) = [].

apply_cfilter(keep, Content) = [Content].

apply_cfilter(elem, celem(E)) = [celem(E)].
apply_cfilter(elem, ctext(_)) = [].

apply_cfilter(text, celem(_)) = [].
apply_cfilter(text, ctext(T)) = [ctext(T)].

apply_cfilter(children, celem(E)) = E ^ content.
apply_cfilter(children, ctext(_)) = [].

apply_cfilter(tag(Tag), celem(E)) = ( E ^ name = Tag -> [celem(E)] ; []).
apply_cfilter(tag(_),   ctext(_)) = [].

apply_cfilter(attr(A), celem(E)) =
    ( find_attribute(A, E ^ attrs, _) ->
        [celem(E)]
    ;
        []
    ).
apply_cfilter(attr(_), ctext(_)) = [].

apply_cfilter(attrval(A), celem(E)) =
    ( find_attribute(A ^ attr_name, E ^ attrs, A) ->
        [celem(E)]
    ;
        []
    ).
apply_cfilter(attrval(_), ctext(_)) = [].

apply_cfilter(F `o` G, Content) =
    condense(map(apply_cfilter(F), apply_cfilter(G, Content))).

apply_cfilter(F ++ G, Content) =
    apply_cfilter(F, Content) ++ apply_cfilter(G, Content).

apply_cfilter(F `with` G, Content) =
    filter(nonempty(G), apply_cfilter(F, Content)).

apply_cfilter(F `without` G, Content) =
    filter(empty(G), apply_cfilter(F, Content)).

apply_cfilter(F `'/>'` G, Content) =
    apply_cfilter(G `o` children `o` F, Content).

apply_cfilter(F `'</'` G, Content) =
    apply_cfilter(F `with` (G `o` children), Content).

apply_cfilter(F `'|>|'` G, Content) =
    apply_cfilter(ife(F, F, G), Content).

apply_cfilter(cat(Fs), Content) =
    condense(map(func(F) = apply_cfilter(F, Content), Fs)).

apply_cfilter(ife(C, T, E), Content) =
    ( nonempty(C, Content) ->
        apply_cfilter(T, Content)
    ;
        apply_cfilter(E, Content)
    ).

apply_cfilter(deep(F), Content) =
    apply_cfilter(F `'|>|'` ife(children, deep(F `o` children), none), Content).

apply_cfilter(deepest(F), Content) =
    % apply_cfilter(deepest(F `o` children) `'|>|'` F, Content).
    apply_cfilter(
        ife(F `o` children, deepest(F `o` children), none) `'|>|'` F,
        Content).

apply_cfilter(multi(F), Content) =
    % apply_cfilter(F ++ multi(F `o` children), Content).
    apply_cfilter(F ++
            ife(F `o` children, multi(F `o` children), none), Content).

apply_cfilter(chip(F), Content) = 
    ( Content = celem(elem(N, As, _)) ->
        [celem(elem(N, As, apply_cfilter(F `o` children, Content)))]
    ;
        [Content]
    ).

apply_cfilter(foldXml(F), Content) =
    apply_cfilter(F `o` chip(foldXml(F)), Content).
    % apply_cfilter(ife(elem, F `o` chip(foldXml(F)), keep), Content).

:- pred find_attribute(string::in,
                list(attribute)::in, attribute::out) is semidet.
find_attribute(Name, [Attr | Attrs], Attribute) :-
    ( Attr ^ attr_name = Name ->
        Attribute = Attr
    ;
        find_attribute(Name, Attrs, Attribute)
    ).

:- pred nonempty(cfilter::in, content::in) is semidet.
nonempty(F, C) :-
    not empty(F, C).
    
:- pred empty(cfilter::in, content::in) is semidet.
empty(F, C) :-
    apply_cfilter(F, C) = [].


%------------------------------------------------------------------------------%

write_element(Stream, Element, !IO) :-
    io__set_output_stream(Stream, OldStream, !IO),
    write_element(Element, !IO),
    io__set_output_stream(OldStream, _, !IO).

write_element(elem(Name, Attrs, Content), !IO) :-
    io__write_string("<", !IO),
    io__write_string(escape(Name), !IO),
    list__foldl(write_attribute, Attrs, !IO),
    ( Content = [] ->
        io__write_string("/>", !IO)
    ;
        io__write_string(">", !IO),
        list__foldl(write_content, Content, !IO),
        io__write_string("</", !IO),
        io__write_string(Name, !IO),
        io__write_string(">", !IO)
    ).

:- pred write_attribute(attribute::in, io::di, io::uo) is det.
write_attribute(attr(Name, Value), !IO) :-
    io__write_string(" ", !IO),
    io__write_string(escape(Name), !IO),
    io__write_string("=""", !IO),
    io__write_string(escape(Value), !IO),
    io__write_string("""", !IO).

:- pred write_content(content::in, io::di, io::uo) is det.
write_content(celem(Element), !IO) :-
    write_element(Element, !IO).
write_content(ctext(Text), !IO) :-
    io__write_string(escape(Text), !IO).

%------------------------------------------------------------------------------%

:- type parser
        --->    parser(
                    stack   :: list(element),
                    current :: element
                ).

:- instance mexpat(parser) where [
    (start(Tag, As, P) = parser(Stack, Elem) :-
        Stack = [P ^ current | P ^ stack],
        Elem = elem(Tag, list__map(to_attr, As), [])
    ),
    (end(_, P0) = P :-
        ( P0 ^ stack = [Head | Stack] ->
            P0 ^ current = elem(Tag, Attrs, ReversedContents),
            C = elem(Tag, Attrs, merge(ReversedContents, [], [])),
                
            Head = elem(T, As, Cs),
            
            P = parser(Stack, elem(T, As, [celem(C) | Cs]))
        ;
            error("stack empty")
        )
    ),
    (cdata(CText, P) = parser(P ^ stack, E) :-
        P ^ current = elem(Tag, Attributes, Content),
        E = elem(Tag, Attributes, [ctext(CText) | Content])
    )
].

    %
    % Given a list of content in reverse order, reverse the
    % list of content while at the same time merging all the
    % consecutive ctext elements into one ctext element.
    %
:- func merge(list(content), list(string), list(content)) = list(content).

merge([], [], L) = L.
merge([], [H | T], L) = [ctext(append_list([H | T])) | L].
merge([ctext(Text) | T], Texts, L) = merge(T, [Text | Texts], L).
merge([celem(Elem) | T], Texts, L) =
    ( Texts = [] ->
        merge(T, [], [celem(Elem) | L])
    ;
        merge(T, [], [celem(Elem), ctext(append_list(Texts)) | L])
    ).
    
parser_init = parser([], elem("dummy_root", [], [])).

get_element(parser([], elem("dummy_root", [], [celem(Elem)])), Elem).

:- func to_attr({string, string}) = attribute.
to_attr({Name, Value}) = attr(Name, Value).

%------------------------------------------------------------------------------%

child_elements(Element) = Children :-
    list__filter_map(
        (pred(Content::in, Child::out) is semidet :-
            Content = celem(Child)
        ), Element ^ content, Children).
        
attribute(Name, As) =
    ( V = As ^ attr(Name) ->
        V
    ;
        func_error("attribute `" ++ Name ++ "' can't be found in " ++
                string(As))
    ).
    
attr(Name, [attr(N, V) | As]) =
    ( Name = N ->
        V
    ;
        attr(Name, As)
    ).

txt(ctext(T)) = T.
txt(celem(E)) =
    func_error("txt(" ++ string(celem(E)) ++ "): not a text content").

el(celem(E)) = E.
el(ctext(T)) =
    func_error("el(" ++ string(ctext(T)) ++ "): not an element content").

serialize(elem(Name, Attrs, Contents)) =
    "<" ++ escape(Name) ++ append_list(list__map(serialize_attribute, Attrs)) ++ ">" ++
        append_list(list__map(serialize_content, Contents)) ++
    "</" ++ escape(Name) ++ ">".

:- func serialize_attribute(attribute) = string.

    serialize_attribute(attr(Key, Value)) =
       " " ++ escape(Key) ++ "=\"" ++ escape(Value) ++ "\"".

:- func serialize_content(content) = string.

serialize_content(ctext(T)) = escape(T).
serialize_content(celem(E)) = serialize(E).

%------------------------------------------------------------------------------%

:- instance xmlize(string) where [
    to_xml(Str) = elem("string", [attr("value", Str)], []),
    from_xml(Elem) =
        ( Elem = elem("string", [attr("value", Str)], []) ->
            Str
        ;
            func_error("from_xml: not a string " ++ string(Elem))
        )
].
:- instance xmlize(int) where [
    to_xml(X) = elem("int", [attr("value", int_to_string(X))], []),
    from_xml(Elem) =
        ( Elem = elem("int", [attr("value", Str)], []) ->
            det_to_int(Str)
        ;
            func_error("from_xml: not an int " ++ string(Elem))
        )
].

:- instance xmlize(list(T)) <= xmlize(T) where
[
    to_xml(List) = elem("list", [], list__map(func(X) = celem(to_xml(X)), List)),
    from_xml(Elem) =
        ( Elem = elem("list", [], Content) ->
            list__map(from_content, Content)
        ;
            func_error("from_xml: not a list " ++ string(Elem))
        )
].

:- func from_content(content) = T <= xmlize(T).

from_content(celem(E)) = from_xml(E).
from_content(ctext(T)) = func_error("from_content: not an element " ++ string(ctext(T))).

%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%
% vim: ft=mercury ts=4 sw=4 et tw=0 wm=0
